-- TODO:
--   - automatic clock infrastructure configuration
--   - clock selection
--   - clock divider
--   - channel delay configuration
--   - serial stream generation (RAM read, overflow management, modes)
--   - temperature somehow
--   - frequency sense
--   - LED outputs

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.FfpgCorePkg.all;
use work.ffpg_wbgen2_pkg.all;

entity FfpgCore is
    generic (
        CLK_FREQ: natural -- input clock frequency in MHz
    );
    port (
        -- Wishbone connection
        rst_n_i: in std_logic;
        clk_sys_i: in std_logic;
        wb_adr_i: in std_logic_vector(31 downto 0);
        wb_dat_i: in std_logic_vector(31 downto 0);
        wb_dat_o: out std_logic_vector(31 downto 0);
        wb_cyc_i: in std_logic;
        wb_sel_i: in std_logic_vector(3 downto 0);
        wb_stb_i: in std_logic;
        wb_we_i: in std_logic;
        wb_ack_o: out std_logic;
        wb_stall_o: out std_logic;
        -- FMC interface
        TriggerDac_o: out t_FfpgAd5600Interface := c_FfpgAd5600InterfaceInitValue;
        VcxoDac_o: out t_FfpgAd5600Interface := c_FfpgAd5600InterfaceInitValue;
        ch1_out_en: out std_logic;
        ch2_out_en: out std_logic
    );
end entity;

architecture syn of FfpgCore is

	--! computes least possible width of the vector to store value X
	-- http://stackoverflow.com/a/12751341/615627
	function f_log2(x: positive) return natural is
		variable i: natural;
	begin
		i := 0;
		while (2**i < x) loop
			i := i + 1;
		end loop;
		return i;
	end function;

    function calculateSclkDivsel (CLK_FREQ, DAC_FREQ_MAX: natural) return std_logic_vector is
        variable result: integer;
    begin
        result := f_log2(CLK_FREQ/DAC_FREQ_MAX);
        if result < 3 then
            result := 0;
        else
            result := result - 3;
        end if;
        assert result < 8
            report "Clock frequency is too high for use with module gc_serial_dac!"
            severity failure;
        return std_logic_vector(to_unsigned(result, 3));
    end function;

    constant DAC_FREQ_MAX: natural := 20; -- in MHz
    constant SCLK_DIVSEL: std_logic_vector(2 downto 0) := calculateSclkDivsel(CLK_FREQ, DAC_FREQ_MAX);
    
    constant WB_DATA_WIDTH: positive := 32;
    constant WB_ADDRESS_WIDTH: positive := 32;
    
    signal SerialStreamClock: std_logic;
    signal WbRegsInput: t_ffpg_in_registers;
    signal WbRegsOutput: t_ffpg_out_registers;
    
    signal TriggerDacEn, VcxoDacEn: std_logic;
    
    type t_ChannelMemory is record
        Addr: std_logic_vector(10 downto 0); -- Read address
        Data: std_logic_vector(31 downto 0); -- Read data output
        Rd: std_logic; -- Read strobe input (active high)
    end record;
    
    signal Ch1SetMem: t_ChannelMemory;
    signal Ch1ResMem: t_ChannelMemory;
    signal Ch2SetMem: t_ChannelMemory;
    signal Ch2ResMem: t_ChannelMemory;

    ---- extension to WB slave
    -- registers for LOAD_EXT fields
    signal control_clock_selection: std_logic_vector(1 downto 0);
    signal control_ch1_mode: std_logic_vector(1 downto 0);
    signal control_ch2_mode: std_logic_vector(1 downto 0);
    signal vcxo_register: std_logic_vector(15 downto 0);
    signal clock_divider_hi: std_logic_vector(3 downto 0);
    signal ch1_delay_set: std_logic_vector(9 downto 0);
    signal ch1_delay_reset: std_logic_vector(9 downto 0);
    signal ch2_delay_set: std_logic_vector(9 downto 0);
    signal ch2_delay_reset: std_logic_vector(9 downto 0);
    signal trigger_register: std_logic_vector(15 downto 0);
    signal overflow: std_logic_vector(15 downto 0);
    -- delayed load signals
    signal control_clock_selection_load: std_logic;
    signal control_ch1_mode_load: std_logic;
    signal control_ch2_mode_load: std_logic;
    signal vcxo_register_load: std_logic;
    signal clock_divider_hi_load: std_logic;
    signal ch1_delay_set_load: std_logic;
    signal ch1_delay_reset_load: std_logic;
    signal ch2_delay_set_load: std_logic;
    signal ch2_delay_reset_load: std_logic;
    signal trigger_register_load: std_logic;
    signal overflow_load: std_logic;
    
begin

    cWbSlave: entity work.ffpg_csr(syn)
        port map (
            rst_n_i => rst_n_i,
            clk_sys_i => clk_sys_i,
            wb_adr_i => wb_adr_i(13 downto 0),
            wb_dat_i => wb_dat_i,
            wb_dat_o => wb_dat_o,
            wb_cyc_i => wb_cyc_i,
            wb_sel_i => wb_sel_i,
            wb_stb_i => wb_stb_i,
            wb_we_i => wb_we_i,
            wb_ack_o => wb_ack_o,
            wb_stall_o => wb_stall_o,
            serial_stream_clk_ik => SerialStreamClock,
            ffpg_ch1_set_mem_addr_i => Ch1SetMem.Addr,
            ffpg_ch1_set_mem_data_o => Ch1SetMem.Data,
            ffpg_ch1_set_mem_rd_i => Ch1SetMem.Rd,
            ffpg_ch1_res_mem_addr_i => Ch1ResMem.Addr,
            ffpg_ch1_res_mem_data_o => Ch1ResMem.Data,
            ffpg_ch1_res_mem_rd_i => Ch1ResMem.Rd,
            ffpg_ch2_set_mem_addr_i => Ch2SetMem.Addr,
            ffpg_ch2_set_mem_data_o => Ch2SetMem.Data,
            ffpg_ch2_set_mem_rd_i => Ch2SetMem.Rd,
            ffpg_ch2_res_mem_addr_i => Ch2ResMem.Addr,
            ffpg_ch2_res_mem_data_o => Ch2ResMem.Data,
            ffpg_ch2_res_mem_rd_i => Ch2ResMem.Rd,
            regs_i => WbRegsInput,
            regs_o => WbRegsOutput
        );
    
    process (clk_sys_i) is begin
        if rising_edge(clk_sys_i) then
            control_clock_selection_load <= '0';
            control_ch1_mode_load <= '0';
            control_ch2_mode_load <= '0';
            vcxo_register_load <= '0';
            clock_divider_hi_load <= '0';
            ch1_delay_set_load <= '0';
            ch1_delay_reset_load <= '0';
            ch2_delay_set_load <= '0';
            ch2_delay_reset_load <= '0';
            trigger_register_load <= '0';
            overflow_load <= '0';

            if WbRegsOutput.control_clock_selection_load_o then
                control_clock_selection <= WbRegsOutput.control_clock_selection_o;
                control_clock_selection_load <= '1';
            end if;
            if WbRegsOutput.control_ch1_mode_load_o then
                control_ch1_mode <= WbRegsOutput.control_ch1_mode_o;
                control_ch1_mode_load <= '1';
            end if;
            if WbRegsOutput.control_ch2_mode_load_o then
                control_ch2_mode <= WbRegsOutput.control_ch2_mode_o;
                control_ch2_mode_load <= '1';
            end if;
            if WbRegsOutput.vcxo_register_load_o then
                vcxo_register <= WbRegsOutput.vcxo_register_o;
                vcxo_register_load <= '1';
            end if;
            if WbRegsOutput.clock_divider_hi_load_o then
                clock_divider_hi <= WbRegsOutput.clock_divider_hi_o;
                clock_divider_hi_load <= '1';
            end if;
            if WbRegsOutput.ch1_delay_set_load_o then
                ch1_delay_set <= WbRegsOutput.ch1_delay_set_o;
                ch1_delay_set_load <= '1';
            end if;
            if WbRegsOutput.ch1_delay_reset_load_o then
                ch1_delay_reset <= WbRegsOutput.ch1_delay_reset_o;
                ch1_delay_reset_load <= '1';
            end if;
            if WbRegsOutput.ch2_delay_set_load_o then
                ch2_delay_set <= WbRegsOutput.ch2_delay_set_o;
                ch2_delay_set_load <= '1';
            end if;
            if WbRegsOutput.ch2_delay_reset_load_o then
                ch2_delay_reset <= WbRegsOutput.ch2_delay_reset_o;
                ch2_delay_reset_load <= '1';
            end if;
            if WbRegsOutput.trigger_register_load_o then
                trigger_register <= WbRegsOutput.trigger_register_o;
                trigger_register_load <= '1';
            end if;
            if WbRegsOutput.overflow_load_o then
                overflow <= WbRegsOutput.overflow_o;
                overflow_load <= '1';
            end if;
        end if;
    end process;
    WbRegsInput.control_clock_selection_i <= control_clock_selection;
    WbRegsInput.control_ch1_mode_i <= control_ch1_mode;
    WbRegsInput.control_ch2_mode_i <= control_ch2_mode;
    WbRegsInput.vcxo_register_i <= vcxo_register;
    WbRegsInput.clock_divider_hi_i <= clock_divider_hi;
    WbRegsInput.ch1_delay_set_i <= ch1_delay_set;
    WbRegsInput.ch1_delay_reset_i <= ch1_delay_reset;
    WbRegsInput.ch2_delay_set_i <= ch2_delay_set;
    WbRegsInput.ch2_delay_reset_i <= ch2_delay_reset;
    WbRegsInput.trigger_register_i <= trigger_register;
    WbRegsInput.overflow_i <= overflow;
    
    TriggerDacEn <= trigger_register_load and not WbRegsInput.status_dac_trigger_busy_i;
    VcxoDacEn <= vcxo_register_load and not WbRegsInput.status_dac_vcxo_busy_i;

    cTriggerDac: entity work.gc_serial_dac(syn)
        generic map (
            g_num_data_bits => 16,
            g_num_extra_bits => 8,
            g_num_cs_select => 1,
            g_sclk_polarity => 0
        )
        port map (
            clk_i => clk_sys_i,
            rst_n_i => rst_n_i,
            value_i => trigger_register,
            cs_sel_i(0) => TriggerDacEn,
            load_i => TriggerDacEn,
            sclk_divsel_i => SCLK_DIVSEL,
            dac_cs_n_o(0) => TriggerDac_o.FrameSynchronization_n,
            dac_sclk_o => TriggerDac_o.SerialClock,
            dac_sdata_o => TriggerDac_o.SerialData,
            busy_o => WbRegsInput.status_dac_trigger_busy_i
        );

    cVcxoDac: entity work.gc_serial_dac(syn)
        generic map (
            g_num_data_bits => 16,
            g_num_extra_bits => 8,
            g_num_cs_select => 1,
            g_sclk_polarity => 0
        )
        port map (
            clk_i => clk_sys_i,
            rst_n_i => rst_n_i,
            value_i => vcxo_register,
            cs_sel_i(0) => VcxoDacEn,
            load_i => VcxoDacEn,
            sclk_divsel_i => SCLK_DIVSEL,
            dac_cs_n_o(0) => VcxoDac_o.FrameSynchronization_n,
            dac_sclk_o => VcxoDac_o.SerialClock,
            dac_sdata_o => VcxoDac_o.SerialData,
            busy_o => WbRegsInput.status_dac_vcxo_busy_i
        );

    ch1_out_en <= WbRegsOutput.control_ch1_oe_o;
    ch2_out_en <= WbRegsOutput.control_ch2_oe_o;
    -- WbRegsOutput.clock_divider_lo_o
    
    WbRegsInput.status_clock_infrastructure_busy_i <= 'X';
    WbRegsInput.status_clock_selection_busy_i <= 'X';
    WbRegsInput.status_delay_configuration_busy_i <= 'X';
    WbRegsInput.status_channel_1_oe_i <= WbRegsOutput.control_ch1_oe_o;
    WbRegsInput.status_channel_2_oe_i <= WbRegsOutput.control_ch2_oe_o;
    WbRegsInput.status_channel_1_running_i <= 'X';
    WbRegsInput.status_channel_2_running_i <= 'X';
    WbRegsInput.frequency_i <= (others => 'X');

end architecture;
